// include file to create helixen
// with inspirations from "spring.inc" by Ron Parker

// Main idea: the major radius of the segments is slightly larger
// than CoilRadius, thus keeping the helix from breaking apart
// for large values of CoilSpacing.

// Arguments: CoilRadius, CoilSpacing, WireRadius, Coils, Gamma [degrees].
// 2*gamdeg is the angular length of the Coil pieces before rotation.
// gamma is essentially a quality parameter, 30 deg is probablz good.

#macro Helix(CoilRadius,CoilSpacing,WireRadius,Coils,gamdeg)
// copy args to abbreviated names
# local R=CoilRadius;
# local h=CoilSpacing;
# local wr=WireRadius;
# local n=Coils;
// stepsize control (a smaller value reduces convergence speed and
//  the risk of divergence)
# local damping=1;
// compute half-base angle in radians
# local gam=gamdeg*pi/180;
// q=average y-increment per angle
# local q=h/(2*pi);
// aux. values
# local cg=cos(gam);
# local sg=sin(gam);
// initial values for psi to be approximated
// psi=angle by which the segments will be tilted
# local psiNew=atan2(q,R);
# local done=0;
# local steps=0;
// predeclare aux. variables needed in loop:
# local psi=-1;
# local cp=-1;
# local sp=-1;
# local ga_eff=-1;
# local eq1=-1;
# local deq1=-1;
# local psiabs=-1;

// solve system of eqns:
# while(!done)
#  declare psi=psiNew;
// loop-local aux. values:
#  declare cp=cos(psi);
#  declare sp=sin(psi);
#  declare ga_eff=atan2(sg,cg*cp);
// equation to be solved:
#  declare eq1=R*sg*sp/q-(1-cg*sp*sp)*ga_eff;
// derivative of eq1 wrt. psi:
#  declare deq1=R*sg*cp/q+2*cg*cp*sp*ga_eff-(1-cg*sp*sp)/(1+sg*sg/(cg*cg*cp*cp));
// new approximation (Newton's algorithm):
#  declare psiNew=psi-damping*eq1/deq1;
// The following lines might be unnecessary. I keep them in
// since I think it is better to catch a runaway iteration than to
// find a broken helix after a weeklong rendering:
// fold psi back into a sensible interval
#  declare psiabs=abs(psiNew);
#  while(psiabs>pi)
#   warning "readjusting psi\n"
#   declare damping=0.9*damping;
#   declare psiabs=psiabs-pi;
#  end
// give the folded psi the right sign:
#  #declare psiNew=psiabs*q/abs(q);
#  declare done=(abs(eq1)<1.0e-6);
#  if(steps>98)
#   warning concat("psi=",str(psiNew,8,4),"\n")
#   warning concat("eq1=",str(eq1,8,4),", deq1=",str(deq1,8,4),"\n")
#   error "iterations exceeded\n"
#  end // if
# declare steps=steps+1;
# end // while
# warning concat(str(steps,4,0)," iterations needed\n")
# warning concat("eq1=",str(eq1,8,4),", deq1=",str(deq1,8,4),"\n")
# warning concat("psi=",str(psiNew,8,4),"\n")
# warning concat("gamma=",str(gam,8,4),"\n")
# warning concat("gamma_eff=",str(ga_eff,8,4),"\n")

// built single helix sector
# declare psi=psiNew;
# declare cp=cos(psi);
# declare sp=sin(psi);
# local r=R/(1-cg*sp*sp);
# warning concat("r=",str(r,8,4),"\n")
// The 0.00001 in the lines below avoids problems with identical
// surfaces, for nontransparent media 0 is good enough.
# local Untilted_Helix_Segment=intersection {
  torus { r,wr }
  plane { <cg,0,sg>,0.00001 }
  plane { <-cg,0,sg>,0.00001 }
  translate <0,0,r-R>}
# local Helix_segment=object { Untilted_Helix_Segment rotate z*degrees(psi) }
// dimensions of titled piece:
# declare ga_eff=atan2(sg,cg*cp); // half-angle
# local dh_eff=r*sg*sp;         // half-height
# local nPieces=2*pi*n/(2*ga_eff);
# local i=0;
  merge {
# while(i<nPieces)
   object { Helix_segment
     translate (2*i+1)*dh_eff*y
     rotate degrees(-2*ga_eff)*(i+0.5)*y
}
#  declare i=i+1;
# end // while
} // for union
#end // for macro

//  Usage
//  -----
//  The following code should work:
//  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//  #version 3.1;
//  camera { location <0,3,-8> direction <0,0,1> look_at <0,3,0> }
//  
//  light_source { <0,0,-40> color red 1 green 0 blue 0 }
//  light_source { <0,20,-40> color red 0 green 1 blue 0 }
//  light_source { <0,-20,-40> color red 0 green 0 blue 1 }
//  
//  #include "mkhelix.inc"
//  #declare wplastic=texture {
//    pigment { color rgb <1,1,1> }
//    finish { ambient 0.1 diffuse 0.6 phong 0.25 phong_size 40 }
//  }
//  
//  object {
//    // args: CR,H,WR,N,gamma[degrees]
//    Helix(3,0.3,0.12,20,5)
//    texture { wplastic }
//  }
//  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
